# -*- coding: utf-8 -*-

# from pylint-django

import os
import sys

import pytest

from pylint.test import test_functional


class PylintMongoEngineLintModuleTest(test_functional.LintModuleTest):
    """
        Only used so that we can load this plugin into the linter!
    """

    def __init__(self, test_file):
        super().__init__(test_file)
        self._linter.load_plugin_modules(['pylint_mongoengine'])


def get_tests(input_dir='input', sort=False):
    def _file_name(test):
        return test.base

    HERE = os.path.dirname(os.path.abspath(__file__))
    input_dir = os.path.join(HERE, input_dir)

    suite = []
    for fname in os.listdir(input_dir):
        if fname != '__init__.py' and fname.endswith('.py'):
            suite.append(test_functional.FunctionalTestFile(input_dir, fname))

    return suite


TESTS = get_tests()
TESTS_NAMES = [t.base for t in TESTS]


@pytest.mark.parametrize("test_file", TESTS, ids=TESTS_NAMES)
def test_everything(test_file):
    # copied from pylint.tests.test_functional.test_functional
    LintTest = PylintMongoEngineLintModuleTest(test_file)
    LintTest.setUp()
    LintTest._runTest()


if __name__ == '__main__':
    sys.exit(pytest.main(sys.argv))



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

from pylint_mongoengine.utils import (
    name_is_from_qs,
    node_is_default_qs,
)


class MongoEngineChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'mongoengine-checker'

    msgs = {
        'W6199': ('Placeholder message to prevent disabling of checker',
                  'mongoengine-placeholder',
                  'PyLint does not recognise checkers as being enabled unless'
                  '  they have at least one message which is not fatal...')
    }

    def _called_thru_default_qs(self, node):
        """Checks if an attribute is being accessed throught the default
        queryset manager, ie: MyClass.objects.filter(some='value')"""
        last_child = node.last_child()
        if not last_child:
            return False
        return node_is_default_qs(last_child)

    def check_qs_name(self, node):
        if self._called_thru_default_qs(node) and not name_is_from_qs(
                node.attrname):
            self.add_message('no-member', node=node, args=(
                'QuerySet instance', 'objects', node.attrname, ''))

    @check_messages('no-member')
    def visit_attribute(self, node):
        self.check_qs_name(node)



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from pylint_mongoengine.checkers.mongoengine import MongoEngineChecker
from pylint_mongoengine.suppression import (suppress_qs_decorator_messages,
                                            suppress_fields_attrs_messages,
                                            suppress_fields_messages,
                                            suppress_doc_messages)


def register(linter):
    """Add the needed transformations and supressions.
    """

    linter.register_checker(MongoEngineChecker(linter))
    suppress_qs_decorator_messages(linter)
    suppress_fields_attrs_messages(linter)
    suppress_fields_messages(linter)
    suppress_doc_messages(linter)



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import Mock, MagicMock, patch

import astroid
from pylint_mongoengine import suppression


def test_is_custom_qs_ok():
    funcdef = Mock()
    dec = Mock()
    dec.name = 'queryset_manager'
    funcdef.decorators.get_children.return_value = [None, dec]
    r = suppression._is_custom_qs_manager(funcdef)
    assert r is True


def test_is_custom_qs_no_qs():
    funcdef = Mock()
    funcdef.decorators.get_children.return_value = []
    r = suppression._is_custom_qs_manager(funcdef)
    assert r is False


def test_is_call2custom_manager():
    node = MagicMock()
    node._proxied = MagicMock()
    r = suppression._is_call2custom_manager(node)
    assert r is False


def test_is_custom_manager_attribute_bad_name():
    node = MagicMock()
    node.attrname = 'not_from_qs'
    r = suppression._is_custom_manager_attribute(node)
    assert r is False


@patch.object(suppression, '_is_custom_qs_manager', Mock(return_value=False))
def test_is_custom_manager_attribute_no_qs_manager():
    node = MagicMock()
    node.attrname = 'filter'
    node.get_children.return_value = [MagicMock()]
    r = suppression._is_custom_manager_attribute(node)
    assert r is False


@patch.object(suppression, '_is_custom_qs_manager', Mock(return_value=True))
def test_is_custom_manager_attribute_true():
    node = MagicMock()
    node.attrname = 'filter'
    node.get_children.return_value = [MagicMock()]
    r = suppression._is_custom_manager_attribute(node)
    assert r is True


@patch.object(suppression, 'node_is_embedded_doc', Mock(return_value=True))
@patch.object(suppression, 'node_is_embedded_doc_attr',
              Mock(return_value=True))
def test_is_embedded_doc_attr_true():
    node = MagicMock()
    r = suppression._is_embedded_doc_attr(node)

    assert r is True


@patch.object(suppression, 'node_is_embedded_doc', Mock(return_value=False))
@patch.object(suppression, 'node_is_embedded_doc_attr',
              Mock(return_value=True))
def test_is_embedded_doc_attr_false():
    node = MagicMock()
    r = suppression._is_embedded_doc_attr(node)

    assert r is False


call_of_call = """

def afn(a):
   return lambda: a

afn('asdf')()
"""


def test_is_doc_call_false_call():
    m = astroid.parse(call_of_call)
    body = m.body
    exp = body[1].last_child()
    r = suppression._is_doc_call(exp)
    assert r is False



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from astroid.exceptions import InferenceError
from pylint.checkers.typecheck import TypeChecker, IterableChecker
from pylint.checkers.utils import safe_infer
from pylint_plugin_utils import suppress_message
from pylint_mongoengine.utils import (name_is_from_qs, is_field_method,
                                      node_is_embedded_doc,
                                      node_is_embedded_doc_attr,
                                      node_is_complex_field,
                                      node_is_document,
                                      name_is_from_model,
                                      node_is_default_qs,
                                      name_is_doc_callable)


def _is_custom_qs_manager(funcdef):
    """Checks if a function definition is a queryset manager created
    with the @queryset_manager decorator."""

    decors = getattr(funcdef, 'decorators', None)
    if decors:
        for dec in decors.get_children():
            attrname = 'name' if hasattr(dec, 'name') else 'attrname'
            qs_name = 'queryset_manager'
            try:
                if getattr(dec, attrname) == qs_name:  # pragma no branch
                    return True
            except AttributeError:
                continue

    return False


def _is_call2custom_manager(node):
    """Checks if the call is being done to a custom queryset manager."""
    called = safe_infer(node.func)
    funcdef = getattr(called, '_proxied', None)
    return _is_custom_qs_manager(funcdef)


def _is_custom_manager_attribute(node):
    """Checks if the attribute is a valid attribute for a queryset manager.
    """

    attrname = node.attrname
    if not name_is_from_qs(attrname):
        return False

    for attr in node.get_children():
        inferred = safe_infer(attr)
        funcdef = getattr(inferred, '_proxied', None)
        if _is_custom_qs_manager(funcdef):
            return True

    return False


def _is_embedded_doc_attr(node):
    if node_is_embedded_doc(
            node.last_child()) and node_is_embedded_doc_attr(node):

        return True

    return False


def _is_complex_field_for(node):
    attr = list(node.get_children())[1]
    r = node_is_complex_field(attr)
    return r


def _is_complex_field_compare(node):
    r = False
    _, right = node.ops[0]
    r = node_is_complex_field(right)
    return r


def _is_complex_field_subscript(node):
    attr = node.value
    r = node_is_complex_field(attr)
    return r


def _is_document_field(node):
    """Checks if a node is a valid document field."""
    try:
        inf = next(node.get_children()).inferred()[0]
    except InferenceError:
        return False
    return node_is_document(inf) and name_is_from_model(node.attrname)


def _is_doc_call(node):
    """Checks if node is a valid callable for a document.
    """
    call_node = next(node.get_children())
    call_name = getattr(call_node, 'attrname', None) or getattr(
        call_node, 'name', None)
    if not call_name:
        return False
    # if is a doc call this is a doc
    maybe_doc = node.last_child().last_child()
    if not maybe_doc:
        return False
    try:
        doc_node = maybe_doc.inferred()[0]
    except InferenceError:
        return False

    return node_is_document(doc_node) and name_is_doc_callable(call_name)


def suppress_qs_decorator_messages(linter):
    suppress_message(linter, TypeChecker.visit_call, 'unexpected-keyword-arg',
                     _is_call2custom_manager)
    suppress_message(linter, TypeChecker.visit_call, 'no-value-for-parameter',
                     _is_call2custom_manager)
    suppress_message(linter, TypeChecker.visit_attribute, 'no-member',
                     _is_custom_manager_attribute)


def suppress_fields_attrs_messages(linter):
    suppress_message(linter, TypeChecker.visit_attribute, 'no-member',
                     is_field_method)
    suppress_message(linter, TypeChecker.visit_attribute, 'no-member',
                     _is_embedded_doc_attr)


def suppress_fields_messages(linter):
    suppress_message(linter, IterableChecker.visit_for,
                     'not-an-iterable', _is_complex_field_for)

    suppress_message(linter, TypeChecker.visit_compare,
                     'unsupported-membership-test', _is_complex_field_compare)
    suppress_message(linter, TypeChecker.visit_subscript,
                     'unsupported-assignment-operation',
                     _is_complex_field_subscript)
    suppress_message(linter, TypeChecker.visit_subscript,
                     'unsupported-delete-operation',
                     _is_complex_field_subscript)


def suppress_doc_messages(linter):
    suppress_message(linter, TypeChecker.visit_attribute, 'no-member',
                     _is_document_field)
    suppress_message(linter, TypeChecker.visit_attribute, 'no-member',
                     node_is_default_qs)
    suppress_message(linter, TypeChecker.visit_call, 'not-callable',
                     _is_doc_call)



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import Mock

import astroid

from pylint_mongoengine import utils
from pylint_mongoengine.utils import DOCUMENT_BASES


def test_name_is_from_qs_true():
    attrname = 'filter'
    r = utils.name_is_from_qs(attrname)
    assert r is True


def test_name_is_from_qs_false():
    attrname = 'i_dont_exist'
    r = utils.name_is_from_qs(attrname)
    assert r is False


def test_name_is_from_model_true():
    attrname = 'delete'
    r = utils.name_is_from_model(attrname)
    assert r is True


def test_name_is_from_model_false():
    attrname = 'who_are_you'
    r = utils.name_is_from_model(attrname)
    assert r is False


test_doc = """
from mongomotor import Document

class TestDoc(Document):
    meta = {'inherit': True}


class OtherDoc(TestDoc):
    pass
"""

test_cls = """
class TestCls:
    pass
"""

test_funcdef = """
def bla():
    pass
"""

test_doc_int = """
from mongomotor import Document
from mongomotor.fields import IntField

class Doc(Document):
    intf = IntField()
"""

test_doc_str = """
from mongomotor import Document
from mongomotor.fields import StringField

class Doc(Document):
    strf = StringField()
"""

test_doc_other_str = """
from mongomotor import Document
from mongomotor.fields import StringField

class OtherStringField(StringField):
    pass

class Doc(Document):
    strf = OtherStringField()
"""


def test_node_is_subclass_false():
    module = astroid.parse(test_cls)
    cls = module.body[0]
    r = utils.node_is_subclass(cls, *DOCUMENT_BASES)
    assert r is False


def test_node_is_subclass_true():
    module = astroid.parse(test_doc)
    cls = module.body[2]
    r = utils.node_is_subclass(cls, *DOCUMENT_BASES)
    assert r is True


def test_node_is_subclass_bad_node():
    module = astroid.parse(test_funcdef)
    funcdef = module.body[0]
    r = utils.node_is_subclass(funcdef, *DOCUMENT_BASES)
    assert r is False


def test_node_isinstance_bad_node():
    module = astroid.parse(test_funcdef)
    funcdef = module.body[0]
    r = utils.node_is_instance(funcdef, 'mongomotor.Document')
    assert r is False


def test_node_is_instance_false():
    module = astroid.parse(test_doc_int)
    cls = module.body[2]
    node = list(cls.body[0].get_children())[-1]
    r = utils.node_is_instance(next(node.infer()),
                               'mongoengine.fields.StringField')
    assert r is False


def test_node_is_instance_true():
    module = astroid.parse(test_doc_str)
    cls = module.body[2]
    node = list(cls.body[0].get_children())[-1]
    r = utils.node_is_instance(next(node.infer()),
                               'mongoengine.fields.StringField')
    assert r is True


def test_node_is_instance_inheritance():
    module = astroid.parse(test_doc_other_str)
    cls = module.body[3]
    node = list(cls.body[0].get_children())[-1]
    r = utils.node_is_instance(next(node.infer()),
                               'mongoengine.fields.StringField')
    assert r is True


test_field_method = """
from mongomotor import Document
from mongomotor.fields import StringField

class Doc(Document):
    strf = StringField()

    def do_stuff(self):
        return self.strf.split()
"""


def test_is_field_method_true():
    m = astroid.parse(test_field_method)
    attr = m.body[2].last_child().last_child().value.last_child()
    r = utils.is_field_method(attr)
    assert r is True


test_not_field_method = """
from mongomotor import Document
from mongomotor.fields import StringField

class Doc(Document):
    strf = StringField()

    def do_stuff(self):
        return self.strf.bad()
"""


def test_is_field_method_false():
    m = astroid.parse(test_not_field_method)
    attr = m.body[2].last_child().last_child().value.last_child()
    r = utils.is_field_method(attr)
    assert r is False


def test_node_is_doc_field_true():
    m = astroid.parse(test_not_field_method)
    attr = m.body[2].last_child().last_child().value.last_child()
    parent = attr.last_child()
    r = utils.node_is_doc_field(parent)

    assert r is True


test_no_doc_field = """
from mongomotor import Document

class Doc(Document):
    something = 'bla'

    def do_stuff(self):
        return self.something.split()

"""


def test_node_is_doc_field_false():
    m = astroid.parse(test_no_doc_field)
    attr = m.body[1].last_child().last_child().value.last_child()
    parent = attr.last_child()
    r = utils.node_is_doc_field(parent)

    assert r is False


def test_node_is_doc_field_no_infer(mocker):
    mocker.patch.object(utils, 'safe_infer', Mock(return_value=False))
    m = astroid.parse(test_not_field_method)
    attr = m.body[2].last_child().last_child().value.last_child()
    parent = attr.last_child()
    r = utils.node_is_doc_field(parent)

    assert r is False


test_doc_embedded_field = """
from mongomotor import Document, EmbeddedDocument
from mongomotor.fields import EmbeddedDocumentField, MapField


class Edoc(EmbeddedDocument):

    def meth(self):
        pass


class ReqEDoc(EmbeddedDocument):

    def meth(self):
        pass


class Doc(Document):
    something = EmbeddedDocumentField(Edoc)
    other = EmbeddedDocumentField(ReqEDoc, required=True)

    def do_bad_stuff(self):
        return self.something.bad()

    def do_stuff(self):
        return self.other.meth()

"""

test_doc_complex_field = """
from mongomotor import Document
from mongomotor.fields import MapField, IntField

class ComplexDoc(Document):
    complex_field = MapField()
    not_complex = IntField()

    def fn(self):
        self.complex_field
        self.not_complex

"""


def test_node_is_embedded_doc():
    m = astroid.parse(test_doc_embedded_field)
    attr = m.body[4].last_child().last_child().value.last_child()
    parent = attr.last_child()
    r = utils.node_is_embedded_doc(parent)

    assert r is True


def test_get_node_parent_class():
    m = astroid.parse(test_doc_embedded_field)
    attr = m.body[4].last_child().last_child().value.last_child()
    parent = attr.last_child()
    cls = utils.get_node_parent_class(parent)

    assert isinstance(cls, utils.ClassDef)


def test_get_field_definition():
    m = astroid.parse(test_doc_embedded_field)
    attr = m.body[4].last_child().last_child().value.last_child()
    parent = attr.last_child()
    definition = utils.get_field_definition(parent)
    real_def = m.body[4].body[1]

    assert definition is real_def


def test_get_field_embedded_doc():
    m = astroid.parse(test_doc_embedded_field)
    attr = m.body[4].last_child().last_child().value.last_child()
    parent = attr.last_child()
    cls = utils.get_field_embedded_doc(parent)
    real_cls = m.body[3]

    assert cls is real_cls


def test_node_is_embedded_doc_attr_false():
    m = astroid.parse(test_doc_embedded_field)
    attr = m.body[4].body[2].last_child().last_child().last_child()
    r = utils.node_is_embedded_doc_attr(attr)

    assert r is False


def test_node_is_embedded_doc_attr_true():
    m = astroid.parse(test_doc_embedded_field)
    attr = list(m.body[4].body[3].last_child().last_child().get_children())[0]
    r = utils.node_is_embedded_doc_attr(attr)

    assert r is True


def test_node_is_complex_field_true():
    m = astroid.parse(test_doc_complex_field)
    attr = list(m.body[2].body[2].body[0].get_children())[0]
    r = utils.node_is_complex_field(attr)

    assert r is True


def test_node_is_complex_field_false():
    m = astroid.parse(test_doc_complex_field)
    attr = list(m.body[2].body[2].body[1].get_children())[0]

    r = utils.node_is_complex_field(attr)

    assert r is False


def test_node_is_doc_instance_true():
    m = astroid.parse(test_no_doc_field)
    attr = m.body[1].last_child().last_child().value.last_child().last_child()
    parent = attr.last_child().inferred()[0]
    r = utils.node_is_doc_instance(parent)
    assert r is True



# -*- coding: utf-8 -*-

from astroid.bases import Instance
from astroid.nodes import ClassDef
from astroid.exceptions import InferenceError
from mongoengine import Document
from mongoengine.queryset import QuerySet
from pylint.checkers.utils import safe_infer

DOCUMENT_BASES = ('mongomotor.Document',
                  'mongoengine.Document',
                  'mongoengine.document.Document',
                  'mongomotor.document.Document')


FIELD_TYPES = {'mongoengine.fields.StringField': str(),
               'mongoengine.fields.IntField': int(),
               'mongoengine.fields.FloatField': float(),
               'mongoengine.fields.ListField': list(),
               'mongoengine.fields.DictField': dict()}


qs_names = set()
document_names = set()


def name_is_from_qs(attrname):
    global qs_names  # pylint: disable=global-statement

    if not qs_names:
        qs_names = set(dir(QuerySet))

    return attrname in qs_names


def name_is_from_model(attrname):
    global document_names  # pylint: disable=global-statement

    if not document_names:
        document_names = set(dir(Document))
        # these attributes are dynamically created
        document_names.add('id')
        document_names.add('objects')
        document_names.add('DoesNotExist')
        document_names.add('MultipleObjectsReturned')

    return attrname in document_names


def name_is_doc_callable(attrname):
    """Checks if a name is a callable in a document"""
    is_call = callable(getattr(Document, attrname, None))
    return name_is_from_model and is_call


def node_is_subclass(cls, *subclass_names):
    """Checks if cls node has parent with subclass_name."""
    if not isinstance(cls, (ClassDef, Instance)):
        return False

    for base_cls in cls.bases:
        try:
            for inf in base_cls.inferred():  # pragma no branch
                if inf.qname() in subclass_names:
                    return True

                if inf != cls and node_is_subclass(  # pragma no branch
                        inf, *subclass_names):
                    # check up the hierarchy in case we are a subclass of
                    # a subclass of a subclass ...
                    return True
        except InferenceError:  # pragma no cover
            continue

    return False


def node_is_instance(inst, *cls_names):
    if not isinstance(inst, Instance):
        return False

    if inst.qname() in cls_names:
        return True
    if node_is_subclass(inst, *cls_names):
        return True
    return False


def node_is_document(node):
    """Checks if a node is a mongoengine document class.
    """
    return node_is_subclass(node, *DOCUMENT_BASES)


def node_is_doc_instance(node):
    return node_is_instance(node, *DOCUMENT_BASES)


def node_is_doc_field(node):
    inferred = safe_infer(node)
    if not inferred:
        return False

    return node_is_instance(inferred, 'mongoengine.base.fields.BaseField',
                            'mongoengine.base.fields.ComplexBaseField')


def node_is_default_qs(node):
    # the default qs manager is called 'objects', we check for it here
    attrname = getattr(node, 'attrname', None)
    if attrname != 'objects':
        return False

    base_cls = node.last_child()
    for cls in base_cls.inferred():
        if node_is_document(cls):
            return True

    return False


def is_field_method(node):
    """Checks if a call to a field instance method is valid. A call is
    valid if the call is a method of the underlying type. So, in a StringField
    the methods from str are valid, in a ListField the methods from list are
    valid and so on..."""
    name = node.attrname
    parent = node.last_child()
    inferred = safe_infer(parent)

    if not inferred:
        return False

    for cls_name, inst in FIELD_TYPES.items():
        if node_is_instance(inferred, cls_name) and hasattr(inst, name):
            return True

    return False


def get_node_parent_class(node):
    """Supposes that node is a mongoengine field in a class and tries to
    get its parent class"""

    while node.parent:  # pragma no branch
        if isinstance(node, ClassDef):
            return node

        node = node.parent


def node_is_embedded_doc(node):
    cls_name = 'mongoengine.fields.EmbeddedDocumentField'
    inferred = safe_infer(node)
    if not inferred:
        return False

    return node_is_instance(inferred, cls_name)


def get_field_definition(node):
    """"node is a class attribute that is a mongoengine. Returns
     the definition statement for the attribute
    """

    name = node.attrname
    cls = get_node_parent_class(node)
    definition = cls.lookup(name)[1][0].statement()
    return definition


def get_field_embedded_doc(node):
    """Returns de ClassDef for the related embedded document in a
    embedded document field."""

    definition = get_field_definition(node)
    cls_name = list(definition.last_child().get_children())[1]
    cls = next(cls_name.infer())
    return cls


def node_is_embedded_doc_attr(node):
    """Checks if a node is a valid field or method in a embedded document.
    """
    embedded_doc = get_field_embedded_doc(node.last_child())
    name = node.attrname
    try:
        r = bool(embedded_doc.lookup(name)[1][0])
    except (IndexError, TypeError):
        r = False

    return r


def node_is_complex_field(node):
    cls_name = 'mongoengine.base.fields.ComplexBaseField'
    inferred = safe_infer(node)
    r = node_is_subclass(inferred, cls_name)
    return r



# -*- coding: utf-8 -*-
# Copyright 2018 Juca Crispim <juca@poraodojuca.net>

# This file is part of pylint-mongoengine.

# pylint-mongoengine is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pylint-mongoengine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pylint-mongoengine. If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import Mock, patch, MagicMock

from pylint_mongoengine.checkers.mongoengine import MongoEngineChecker


class TestMongoEngineChecker:

    def setup_method(self, test_method):
        linter = Mock()
        self.checker = MongoEngineChecker(linter)

    def test_called_thru_default_qs_no_last_child(self):
        node = Mock()
        node.last_child.return_value = None
        r = self.checker._called_thru_default_qs(node)
        assert r is False

    def test_called_thru_default_qs_no_objects(self):
        node = Mock()
        node.last_child.return_value.attrname = 'something'
        r = self.checker._called_thru_default_qs(node)
        assert r is False

    @patch('pylint_mongoengine.utils.node_is_document',
           Mock(return_value=False))
    def test_called_thru_default_qs_no_qs(self):
        node = Mock()
        node.last_child.return_value.attrname = 'objects'
        base_cls = Mock()
        base_cls.inferred.return_value = [Mock()]
        node.last_child.return_value.last_child.return_value = base_cls
        r = self.checker._called_thru_default_qs(node)
        assert r is False

    @patch('pylint_mongoengine.checkers.mongoengine.node_is_default_qs',
           Mock(return_value=True))
    def test_called_thru_default_qs_ok(self):
        node = Mock()
        node.last_child.return_value.attrname = 'objects'
        base_cls = Mock()
        base_cls.inferred.return_value = [Mock()]
        node.last_child.return_value.last_child.return_value = base_cls
        r = self.checker._called_thru_default_qs(node)
        assert r is True

    def test_visit_attribute_not_called_thru_qs(self):
        self.checker._called_thru_default_qs = Mock(return_value=False)
        self.checker.add_message = Mock()
        self.checker.visit_attribute(Mock())

        assert self.checker.add_message.called is False

    @patch('pylint_mongoengine.utils.name_is_from_qs',
           Mock(return_value=False))
    def test_visit_attribute_not_qs_name(self):
        self.checker._called_thru_default_qs = Mock(return_value=True)
        self.checker.add_message = Mock()
        self.checker.visit_attribute(Mock())

        assert self.checker.add_message.called is True

    @patch('pylint_mongoengine.checkers.mongoengine.name_is_from_qs',
           Mock(return_value=True))
    def test_visit_attribute_qs_name(self):
        self.checker._called_thru_default_qs = Mock(return_value=True)
        self.checker.add_message = Mock()
        self.checker.visit_attribute(Mock())

        assert self.checker.add_message.called is False



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

DESCRIPTION = 'A Pylint plugin to help it understand MongoEngine'

setup(
    name='pylint-mongoengine',
    url='https://github.com/jucacrispim/pylint-mongoengine',
    author='Juca Crispim',
    author_email='juca@poraodojuca.net',
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    version='0.4.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'pylint>=2.0', 'pylint-plugin-utils>=0.5'
    ],
    license='GPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Topic :: Software Development :: Quality Assurance',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords=['pylint', 'mongoengine', 'plugin'],
    zip_safe=False,
)
